/***************************************************************************
                          binarytree.cpp  -  description
                             -------------------
    begin                : Wed Jul 13 2005
    copyright            : (C) 2005 by Michael Rosskopf
    email                : rosskopf@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "binarytree.h"

// Constructors and destructors
BinaryTree::BinaryTree(){
  leafNumber = 0;
  value = 0;
  leftChild = 0;
  rightChild = 0;
  edgeLengthToLeftChild = 0;
  edgeLengthToRightChild = 0;
  edgeLengthToLeftChildString = 0;
  treeStringLeftChild = 0;
  edgeLengthToRightChildString = 0;
  treeStringRightChild = 0;
}
BinaryTree::~BinaryTree(){
  if(leftChild != 0){
    leftChild->~BinaryTree();
    delete leftChild;
  }
  if(rightChild != 0){
    rightChild->~BinaryTree();
    delete rightChild;
  }
  if(edgeLengthToLeftChildString != 0) delete edgeLengthToLeftChildString;
  if(treeStringLeftChild != 0) delete treeStringLeftChild;
  if(edgeLengthToRightChildString != 0) delete edgeLengthToRightChildString;
  if(treeStringRightChild != 0) delete treeStringRightChild;
}

// getLeafNumber: Returns the leaf number. Returns '0' for an inner node
long BinaryTree::getLeafNumber(){
  return leafNumber;
}

// getValue: Returns the stored value (it is used here to describe the gene expression)
double BinaryTree::getValue(){
  return value;
}

// getLeftChild: Returns the left child. Returns '0' if left child is empty.
BinaryTree* BinaryTree::getLeftChild(){
  return leftChild;
}

// getRightchild: Returns the right child. Returns '0' if right child is empty.
BinaryTree* BinaryTree::getRightchild(){
  return rightChild;
}

// getEdgeLengthToLeftChild: Returns length of the edge to the left child. Returns '0' if left child is empty.
double BinaryTree::getEdgeLengthToLeftChild(){
  return edgeLengthToLeftChild;
}

// getEdgeLengthToRightChild: Returns length of the edge to the right child. Returns '0' if right child is empty.
double BinaryTree::getEdgeLengthToRightChild(){
  return edgeLengthToRightChild;
}

// getNumberOfLeaves: Returns the number of leaves of the tree.
// If the tree is a leaf, 1 is returned.
long BinaryTree::getNumberOfLeaves(){
  if(leafNumber != 0){
    // Tree is a leaf
    return 1;
  } else {
    long number = 0;
    // If the tree is not a leaf, count the number of leaves of the left and the right subtree
    if(leftChild != 0) number = leftChild->getNumberOfLeaves(); // Recursive call
    if(rightChild != 0) number = number + rightChild->getNumberOfLeaves(); // Recursive call
    return number;
  }
}

// setValue: Set the value of the tree
void BinaryTree::setValue(double _value){
  value = _value;
}

// toString: Output of the tree in Newick-Format
void BinaryTree::toString(){
  if(leafNumber != 0){
    // Tree is a leaf
    cout << leafNumber;
  } else {
    // Tree is not a leaf
    cout << "(";
    if(leftChild != 0){
      // Tree has a left child
      leftChild->toString(); // Recursive call
      cout << ":" << edgeLengthToLeftChild;
      if(rightChild != 0) cout << ",";
    }
    if(rightChild != 0){
      // Tree has a right child
      rightChild->toString(); // Recursive call
      cout << ":" << edgeLengthToRightChild;
    }
    cout << ")";
  }
}

// init: Initialises the tree recursively from a string in Newick-Format
void BinaryTree::init(char* treeString){

  // Initialization
  long treeStringLength = strlen(treeString);
  long numberOfSubtrees = 0;
  long recursionDepth = 0;
  long splitPoint = 0;
  long i,j;
  long treeStringLeftChildStart = 0;
  long treeStringLeftChildEnd = 0;
  long treeStringRightChildStart = 0;
  long treeStringRightChildEnd = 0;

  // Count the number of subtrees. If number == 0, the tree is a leaf.
  for(long i = 0; i < treeStringLength; i++){
    if(treeString[i] == '(') numberOfSubtrees++; // A new open bracket indicates a new subtree
  }

  if(numberOfSubtrees == 0){

    // If the tree is a leaf, it is initialised with a leaf number
    leafNumber = atof(treeString);
    leftChild = 0;
    rightChild = 0;

  } else {

    // If the tree is not a leaf,
    // it is searched for split point between the left and the right subtree in the treeString
    for(i = 0; i < treeStringLength; i++){
      if(treeString[i] == '(') recursionDepth++;
      if(treeString[i] == ')') recursionDepth--;
      if(treeString[i] == ',' && recursionDepth == 1) splitPoint = i;
    }

    // Get the edge length to the left child
    i = splitPoint;
    do{
      i--;
    } while(treeString[i] != ':');
    treeStringLeftChildEnd = i - 1;
    edgeLengthToLeftChildString = new char[splitPoint - treeStringLeftChildEnd - 2];
    for(j = 0; j < splitPoint - treeStringLeftChildEnd - 2; j++){
      edgeLengthToLeftChildString[j] = treeString[treeStringLeftChildEnd + j + 2];
    }
    edgeLengthToLeftChild = atof(edgeLengthToLeftChildString);

    // Cut the treeString for the left subtree and create subtree
    i = -1;
    do{
      i++;
    } while(treeString[i] != '(');
    treeStringLeftChildStart = i + 1;
    treeStringLeftChild = new char[treeStringLeftChildEnd - treeStringLeftChildStart + 1];
    for(j = 0; j < treeStringLeftChildEnd - treeStringLeftChildStart + 1; j++){
      treeStringLeftChild[j] = treeString[treeStringLeftChildStart + j];
    }
    leftChild = new BinaryTree();
    leftChild->init(treeStringLeftChild); // Recursive call to initialise the left subtree

    // Get the edge length to the right child
    i = treeStringLength;
    do{
      i--;
    } while(treeString[i] != ':');
    treeStringRightChildEnd = i - 1;
    edgeLengthToRightChildString = new char[treeStringLength - treeStringRightChildEnd - 2];
    for(j = 0; j < treeStringLength - treeStringRightChildEnd - 2; j++){
      edgeLengthToRightChildString[j] = treeString[treeStringRightChildEnd + j + 2];
    }
    edgeLengthToRightChild = atof(edgeLengthToRightChildString);

    // Cut the treeString for the right subtree
    treeStringRightChildStart = splitPoint + 1;
    treeStringRightChild = new char[treeStringRightChildEnd - treeStringRightChildStart + 1];
    for(j = 0; j < treeStringRightChildEnd - treeStringRightChildStart + 1; j++){
      treeStringRightChild[j] = treeString[treeStringRightChildStart + j];
    }
    rightChild = new BinaryTree();
    rightChild->init(treeStringRightChild); // Recursive call to initialise the right subtree
  }
}

// evolution: Simulates evolution along the tree.
// Simulated gene expression values are stored in the leaves.
// _value is the gene expression level in the root of the tree.
// theta and sigmaM are model parameters.
// gene[] stores the gene expression values in the leaves after the simulation.
void BinaryTree::evolution(double _value,double theta,double sigmaM,double gene[],RandomNumberGenerator *rnd){

  // Initialization
  value = _value;
  double mutationEffect = 0.0;
  int numberOfMutations = 0;
  int i;

  // If the tree is a leave, the result is stored
  if(leafNumber != 0){
    gene[leafNumber - 1] = value;
  } else {
    // If the tree has a left child, evolution is corresponding to the branch to the left child
    if(leftChild != 0){
      numberOfMutations = rnd->getPoissonDistributedNumber(theta * edgeLengthToLeftChild);
      mutationEffect = value;
      for(i = 0; i < numberOfMutations; i++){
        mutationEffect = mutationEffect + (rnd->getNormalDistributedNumber(0.0,sigmaM));
      }
      leftChild->evolution(mutationEffect,theta,sigmaM,gene,rnd); // recursive call
    }
    // If the tree has a right child, evolution is corresponding to the branch to the right child
    if(rightChild != 0){
      numberOfMutations = rnd->getPoissonDistributedNumber(theta * edgeLengthToRightChild);
      mutationEffect = value;
      for(i = 0; i < numberOfMutations; i++){
        mutationEffect = mutationEffect + (rnd->getNormalDistributedNumber(0.0,sigmaM));
      }
      rightChild->evolution(mutationEffect,theta,sigmaM,gene,rnd); // recursive call
    }
  }
}
