/***************************************************************************
                          controller.cpp  -  description
                             -------------------
    begin                : Mon Aug 7 2006
    copyright            : (C) 2006 by 
    email                : rosskopf@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "controller.h"

// Constructors and destructors
Controller::Controller(){

  // Create an object for the program parameters which loads the config.txt
  programParameters = new ProgramParameters();
  
  // Create an object for math functions
  ms = new MathFunctions();
  
  // Create a random number generator
  rnd = new RandomNumberGenerator();
  rnd->init();

  // Check program parameters
  bool valid = true;
  valid = programParameters->checkProgramParameters();

  // Initialize data and load if a file is used
  data = new Data();
  if(valid && programParameters->getMode() == 2){
    cout << "\nLoad data ..." << endl;
    data->load(programParameters->getDataTableFilename());
    // Check whether the data set contains the samples
    valid = data->dataCheck(programParameters);
  }

  // Call of the selected tool if program parameters are valid
  if(valid){
    if(programParameters->getTool() == 1) performBayesianMutationDetection();
    if(programParameters->getTool() == 2) performTajimaTypeTest();
  }
}
Controller::~Controller(){
  if(programParameters != 0) delete programParameters;
  if(data != 0) delete data;
  if(ms != 0) delete ms;
  if(rnd != 0) delete rnd;
}


// Implementation of the Bayesian mutation detection method
void Controller::performBayesianMutationDetection(){

  // Initialization
  int sample1; // Index number of sample 1 in the data set
  int sample2; // Index number of sample 2 in the data set
  double expression; // Buffer for gene expression while simulation
  int numberOfMutations; // Buffer for number of mutation events from the Possion process
  long* numberOfMutationsResults; // Stores the results of the method for each gene
  double* maximalPosteriorProbabilityList; // Stores the posterior probabilities corresponding to the results
  long* mutationNumberCounterByPosteriorProbability = new long[programParameters->getMaxMut() + 1]; // Stores statistic of the number of mutations
  double* probabilityForAGivenNumberOfMutationsEvents = new double[programParameters->getMaxMut() + 1]; // Stores statistic of the number of mutations
  double meanNumberOfMutationsPerGeneByPosteriorProbability; // Mean number of mutations for the whole data set
  double posteriorProbability = 0.0; // Buffer
  double maximalPosteriorProbability = 0.0; // Buffer
  double probabilityForDifferenceByNMutations = 0.0;
  long double factorial;
  double normalizingConstant;
  long mutationsByPosteriorProbability = -1.0; // Buffer
  double difference = 0.0; // Buffer
  double buffer; // Buffer

  for(i = 0; i <= programParameters->getMaxMut(); i++){
    mutationNumberCounterByPosteriorProbability[i] = 0;
  }
  
  // Generate gene expression data
  if(programParameters->getMode() == 1){
    cout << "\nGenerate data ..." << endl;
    data->init(programParameters->getSample1Size() + programParameters->getSample2Size(),programParameters->getNumberOfSimulations(),2);
    data->setSampleStart(0,0);
    data->setSampleSize(0,programParameters->getSample1Size());
    data->setSampleStart(1,programParameters->getSample1Size());
    data->setSampleSize(1,programParameters->getSample2Size());
    for(i = 0; i < data->getSampleSize(0); i++){
      data->setLabel(i,1);
    }
    for(i = 0; i < data->getSampleSize(1); i++){
      data->setLabel(i + data->getSampleStart(1),2);
    }
    for(i = 0; i < data->getNumberOfGenes(); i++){
      expression = 0.0;
      // Create synthetic array data for the first group
      for(j = 0; j < data->getSampleSize(0); j++){
        data->setGene(j,i,rnd->getNormalDistributedNumber(0.0,programParameters->getSigmaE()));
      }
      // Simulate the mutation between the groups
      if(programParameters->getFixedMut() < 0){
        numberOfMutations = rnd->getPoissonDistributedNumber(programParameters->getD());
      } else {
        numberOfMutations = programParameters->getFixedMut();
      }
      for(j = 0; j < numberOfMutations; j++){
        expression = expression + rnd->getNormalDistributedNumber(0.0,programParameters->getSigmaM());
      }
      // Create synthetic array data for the second group
      for(j = 0; j < data->getSampleSize(1); j++){
        data->setGene(j + data->getSampleStart(1),i,expression + rnd->getNormalDistributedNumber(0.0,programParameters->getSigmaE()));
      }
    }

    // Save synthetic data
    data->save("_datatable_synthetic_microarray.txt");

    sample1 = 0;
    sample2 = 1;
  }

  // Full test: Load gene expression data (in this case data has been loaded yet)
  if(programParameters->getMode() == 2){
    // Redirect the sample numbers depending on the parameter file
    sample1 = programParameters->getSample1() - 1;
    sample2 = programParameters->getSample2() - 1;
  }

  // Initialize array for the most likely number of mutations per gene
  numberOfMutationsResults = new long[data->getNumberOfGenes()];
  maximalPosteriorProbabilityList = new double[data->getNumberOfGenes()];

  // Calculate probability for 0, 1, ..., n mutations events
  for(numberOfMutations = 0; numberOfMutations <= programParameters->getMaxMut(); numberOfMutations++){
    factorial = 1;
    for(j = 1; j <= numberOfMutations; j++){
      factorial = factorial * j;
    }
    probabilityForAGivenNumberOfMutationsEvents[numberOfMutations] = (pow(2.718281828,-programParameters->getD()) * pow(programParameters->getD(),numberOfMutations)) / factorial;
  }

  // Bayesian mutation detection method
  cout << "\nPerform mutation detection method ..." << endl;

  for(i = 0; i < data->getNumberOfGenes(); i++){

    mutationsByPosteriorProbability = -1.0;
    maximalPosteriorProbability = 9999999999.0;

    for(numberOfMutations = 0; numberOfMutations <= programParameters->getMaxMut(); numberOfMutations++){

      // Calculate the posterior probability for a given gene expression difference
      posteriorProbability = 0.0;
      for(j = 0; j < data->getSampleSize(sample1); j++){
        for(k = 0; k < data->getSampleSize(sample2); k++){
          difference = data->getGene(data->getSampleStart(sample1) + j,i) - data->getGene(data->getSampleStart(sample2) + k,i);

          if(programParameters->getSigmaE() != 0.0){
          // Normal case with non-mutational effects

	    // Calculate probability that the observed gene expression difference has been caused by the given number of mutations
            probabilityForDifferenceByNMutations = (1.0 / ((pow(2.0 * PI,0.5) * pow(numberOfMutations * pow(programParameters->getSigmaM(),2.0) + 2.0 * pow(programParameters->getSigmaE(),2.0),0.5)))) * pow(2.718281828,-pow(difference,2.0) / (2.0 * (numberOfMutations * pow(programParameters->getSigmaM(),2.0) + 2.0 * pow(programParameters->getSigmaE(),2.0))));
 
            // Calculate probability for the observed gene expression difference
            normalizingConstant = 0.0;
            for(l = 0; l <= programParameters->getMaxMut(); l++){
              buffer = (1.0 / (pow(2.0 * PI,0.5) * pow(l * pow(programParameters->getSigmaM(),2.0) + 2.0 * pow(programParameters->getSigmaE(),2.0),0.5))) * pow(2.718281828,-pow(difference,2.0) / (2.0 * (l * pow(programParameters->getSigmaM(),2.0) + 2.0 * pow(programParameters->getSigmaE(),2.0))));
              normalizingConstant = normalizingConstant + probabilityForAGivenNumberOfMutationsEvents[l] * buffer;
            }
            
	    // Calculate posterior probability
            posteriorProbability = posteriorProbability - log((probabilityForAGivenNumberOfMutationsEvents[numberOfMutations] * probabilityForDifferenceByNMutations) / normalizingConstant);

          } else {
          // Special case if the non-mutational effects have no impact (sigmaE == 0)

            if(numberOfMutations == 0){
              if(difference != 0.0) posteriorProbability = 9999999999.0;
            }
            if(numberOfMutations > 0){
              if(difference == 0.0){
                posteriorProbability = 9999999999.0;
              }
              if(difference != 0.0){
                // Calculate probability that the observed gene expression difference has been caused by the given number of mutations
                probabilityForDifferenceByNMutations = (1.0 / ((pow(2.0 * PI,0.5) * pow(numberOfMutations * pow(programParameters->getSigmaM(),2.0),0.5)))) * pow(2.718281828,-pow(difference,2.0) / (2.0 * (numberOfMutations * pow(programParameters->getSigmaM(),2.0))));
 
                // Calculate probability for the observed gene expression difference
                normalizingConstant = 0.0;
                for(l = 1; l <= programParameters->getMaxMut(); l++){
                  buffer = (1.0 / (pow(2.0 * PI,0.5) * pow(l * pow(programParameters->getSigmaM(),2.0),0.5))) * pow(2.718281828,-pow(difference,2.0) / (2.0 * (l * pow(programParameters->getSigmaM(),2.0))));
                  normalizingConstant = normalizingConstant + probabilityForAGivenNumberOfMutationsEvents[l] * buffer;
                }

                posteriorProbability = posteriorProbability - log((probabilityForAGivenNumberOfMutationsEvents[numberOfMutations] * probabilityForDifferenceByNMutations) / normalizingConstant);
              }
            }
          }
        }
      }
      
      // Has been found a more likely number of events?
      if(posteriorProbability < maximalPosteriorProbability){
        mutationsByPosteriorProbability = numberOfMutations;
        maximalPosteriorProbability = posteriorProbability;
      }
    }
    
    // Count results for result statistics
    numberOfMutationsResults[i] = mutationsByPosteriorProbability;
    maximalPosteriorProbabilityList[i] = maximalPosteriorProbability;
    mutationNumberCounterByPosteriorProbability[mutationsByPosteriorProbability]++;
  }

  // Output of the results to the screen and to a file
  ofstream fileMainResults("_output_mutation-detection_results.txt");
  cout.precision(4);
  cout << "\n----------------------------------------------------------------" << endl;
  cout << "\nBAYESIAN MUTATION DETECTION RESULTS" << endl;
  cout << "\nMut\tNumber\tPercentage\n" << endl;
  fileMainResults.precision(4);
  fileMainResults << "\nBAYESIAN MUTATION DETECTION RESULTS" << endl;
  fileMainResults << "\nMut\tNumber\tPercentage\n" << endl;
  meanNumberOfMutationsPerGeneByPosteriorProbability = 0.0;
  for(i = 0; i <= programParameters->getMaxMut(); i++){
    cout << i << "\t" << mutationNumberCounterByPosteriorProbability[i] << "\t" << (double(mutationNumberCounterByPosteriorProbability[i]) / double(data->getNumberOfGenes())) * 100.0 << "%" << endl;
    fileMainResults << i << "\t" << mutationNumberCounterByPosteriorProbability[i] << "\t" << (double(mutationNumberCounterByPosteriorProbability[i]) / double(data->getNumberOfGenes())) * 100.0 << "%" << endl;
    meanNumberOfMutationsPerGeneByPosteriorProbability = meanNumberOfMutationsPerGeneByPosteriorProbability + (i * mutationNumberCounterByPosteriorProbability[i]);
  }
  meanNumberOfMutationsPerGeneByPosteriorProbability = meanNumberOfMutationsPerGeneByPosteriorProbability / double(data->getNumberOfGenes());

  cout << "\nMean number of mutations per gene:            " << meanNumberOfMutationsPerGeneByPosteriorProbability << endl;
  cout << "\nPercentage of genes with no mutation:         " << (mutationNumberCounterByPosteriorProbability[0] / double(data->getNumberOfGenes())) * 100.0 << "%" << endl;
  cout << "\nPercentage of genes with at least 1 mutation: " << (1.0 - (mutationNumberCounterByPosteriorProbability[0] / double(data->getNumberOfGenes()))) * 100.0 << "%" << endl;
  fileMainResults << "\nMean number of mutations per gene:            " << meanNumberOfMutationsPerGeneByPosteriorProbability << endl;
  fileMainResults << "\nPercentage of genes with no mutation:         " << (mutationNumberCounterByPosteriorProbability[0] / double(data->getNumberOfGenes())) * 100.0 << "%" << endl;
  fileMainResults << "\nPercentage of genes with at least 1 mutation: " << (1.0 - (mutationNumberCounterByPosteriorProbability[0] / double(data->getNumberOfGenes()))) * 100.0 << "%" << endl;
  cout << "\n----------------------------------------------------------------" << endl;
  cout << "\n";
  if(programParameters->getMode() == 1) cout << "Simulated data was written to \"_datatable_synthetic_microarray.txt\"." << endl;
  cout << "Main results were written to \"_output_mutation-detection_main-results.txt\"." << endl;
  cout << "Results for each gene were written to \"_output_mutation-detection_gene-list.txt\"." << endl;

  ofstream file("_output_mutation-detection_gene-list.txt");
  file << "PS ID\t#Mut\tPP" << endl;
  for(i = 0; i < data->getNumberOfGenes(); i++){
    // Synthetic data
    if(programParameters->getMode() == 1){
      file << i + 1 << "\t" << numberOfMutationsResults[i] << "\t" << maximalPosteriorProbabilityList[i] << endl;
    }
    // Real data
    if(programParameters->getMode() == 2){
      file << data->getProbeSetID(i) << "\t" << numberOfMutationsResults[i] << "\t" << maximalPosteriorProbabilityList[i] << endl;
    }
  }
  file.close();

  delete numberOfMutationsResults;
  delete maximalPosteriorProbabilityList;
  delete mutationNumberCounterByPosteriorProbability;
  delete probabilityForAGivenNumberOfMutationsEvents;
}

// Implementation of the Tajima-type test
void Controller::performTajimaTypeTest(){

  // Initialization
  double confidence = programParameters->getConfidenceInterval();
  double lowerLimit = 0.0;
  double upperLimit = 0.0;
  double meanAlleles = 0.0;
  double meanGeneVariance = 0.0;
  double meanThetaAlleles = 0.0;
  double meanThetaVar = 0.0;
  double meanD = 0.0;
  long numberOfPositiveSelectedGenes = 0.0;
  long numberOfNegativeSelectedGenes = 0.0;

  // Generate null-distribution from simulations on genealogies generated by the coalescent process
  // System call to the ms-program (R. Hudson)
  cout << "\nGenerate trees with ms to estimate the confidence limits in neutral case ..." << endl;
  char mscall[64];
  char buffer[20];
  strcpy(mscall,"./ms ");
  // Mode 1: Perform a null-distribution depending on the specified sample size and population mutation rate
  if(programParameters->getMode() == 1) gcvt(programParameters->getSample1Size(),10,buffer);
  // Mode 2: Perform a null-distribution depending on the data
  if(programParameters->getMode() == 2) gcvt(data->getSampleSize(programParameters->getSample1() - 1),10,buffer);
  strcat(mscall,buffer);
  strcat(mscall," ");
  gcvt(programParameters->getNumberOfSimulations(),10,buffer);
  strcat(mscall,buffer);
  strcat(mscall," -T -t ");
  gcvt(programParameters->getD(),10,buffer);
  strcat(mscall,buffer);
  strcat(mscall," >_treefile.txt");
  system(mscall);
  calcConfidenceLimits(programParameters->getSigmaM(),programParameters->getSigmaE(),confidence,&lowerLimit,&upperLimit);

  // Full Tajima-type test for data
  if(programParameters->getMode() == 2){
    
    // Initialise arrays depending on the loaded data set
    long numberOfGenes = data->getNumberOfGenes();
    ThetaAllelesEstimator* se = new ThetaAllelesEstimator();
    double* alleles = new double[numberOfGenes];
    double* geneVariance = new double[numberOfGenes];
    double* thetaAlleles = new double[numberOfGenes];
    double* thetaVar = new double[numberOfGenes];
    double* Delta = new double[numberOfGenes];
    double* vector = new double[numberOfGenes];

    // Calculation of the Delta-values from the data
    for(long i = 0; i < numberOfGenes; i++){
      geneVariance[i] = data->calcGeneVariance(1,i);
      thetaVar[i] = (2.0 * (geneVariance[i] - pow(programParameters->getSigmaE(),2.0))) / pow(programParameters->getSigmaM(),2.0);
      alleles[i] = data->calcNumberOfAlleles(1,i,programParameters->getSigmaM(),programParameters->getSigmaE());
      thetaAlleles[i] = se->estimateThetaFromNumberOfAlleles(data->getSampleSize(programParameters->getSample1() - 1),alleles[i]);
      Delta[i] = thetaVar[i] - thetaAlleles[i];
      meanGeneVariance = meanGeneVariance + geneVariance[i];
      meanThetaVar = meanThetaVar + thetaVar[i];
      meanAlleles = meanAlleles + alleles[i];
      meanThetaAlleles = meanThetaAlleles + thetaAlleles[i];
      meanD = meanD + Delta[i];
    }
    meanGeneVariance = meanGeneVariance / numberOfGenes;
    meanThetaVar = meanThetaVar / numberOfGenes;
    meanAlleles = meanAlleles / numberOfGenes;
    meanThetaAlleles = meanThetaAlleles / numberOfGenes;
    meanD = meanD / numberOfGenes;

    // Write the theta estimates to a file
    ofstream writeFileDDistribution("_output_tajima-type-test_Delta-dist_real-data.txt");
    for(long i = 0; i < numberOfGenes; i++){
      writeFileDDistribution << Delta[i] << endl;
    }
    writeFileDDistribution.close();

    // Write Probe set IDs of directional selected genes to a file (Delta < lower limit)
    ofstream writeFilePositiveSelectedGenes("_output_tajima-type-test_directional_selected_genes.txt");
    for(long i = 0; i < numberOfGenes; i++){
      if(Delta[i] < lowerLimit){
        writeFilePositiveSelectedGenes << data->getProbeSetID(i) << "\t" << Delta[i] << endl;
        numberOfPositiveSelectedGenes++;
      }
    }
    writeFilePositiveSelectedGenes.close();

    // Write Probe set IDs of balancing selected genes to a file  (Delta > upper limit)
    ofstream writeFileNegativeSelectedGenes("_output_tajima-type-test_balancing_selected_genes.txt");
    for(long i = 0; i < numberOfGenes; i++){
      if(Delta[i] > upperLimit){
        writeFileNegativeSelectedGenes << data->getProbeSetID(i) << "\t" << Delta[i] << endl;
        numberOfNegativeSelectedGenes++;
      }
    }
    writeFileNegativeSelectedGenes.close();

    // Output
    ofstream writeFileMainResults2("_output_tajima-type-test_results.txt",ios::app);
    cout << "\nRESULTS FOR THE DATA SET" << endl;
    writeFileMainResults2 << "\nRESULTS FOR THE DATA SET" << endl;

    // Output of the gene variance
    cout << "\nGene variance:" << endl;
    writeFileMainResults2 << "\nGene variance:" << endl;
    for(long i = 0; i < numberOfGenes ; i++){
      vector[i] = geneVariance[i];
    }
    ms->sort(vector,0,numberOfGenes - 1);
    ms->calcMoments(vector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
    cout << "Mean: " << average << endl;
    writeFileMainResults2 << "Mean: " << average << endl;
    cout << confidence * 100.0 << "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    cout << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
    writeFileMainResults2 << confidence * 100.0<< "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    writeFileMainResults2 << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

    // Output of the number of alleles
    cout << "\nNumber of alleles:" << endl;
    writeFileMainResults2 << "\nNumber of alleles:" << endl;
    for(long i = 0; i < numberOfGenes ; i++){
      vector[i] = alleles[i];
    }
    ms->sort(vector,0,numberOfGenes - 1);
    ms->calcMoments(vector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
    cout << "Mean: " << average << endl;
    writeFileMainResults2 << "Mean: " << average << endl;
  cout << confidence * 100.0 << "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    cout << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
    writeFileMainResults2 << confidence * 100.0<< "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    writeFileMainResults2 << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

    // Output of theta_var
    cout << "\ntheta_var:" << endl;
    writeFileMainResults2 << "\ntheta_var:" << endl;
    for(long i = 0; i < numberOfGenes ; i++){
      vector[i] = thetaVar[i];
    }
    ms->sort(vector,0,numberOfGenes - 1);
    ms->calcMoments(vector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
    cout << "Mean: " << average << endl;
    writeFileMainResults2 << "Mean: " << average << endl;
    cout << confidence * 100.0 << "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    cout << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
    writeFileMainResults2 << confidence * 100.0<< "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    writeFileMainResults2 << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

    // Output of theta_alleles
    cout << "\ntheta_alleles:" << endl;
    writeFileMainResults2 << "\ntheta_alleles:" << endl;
    for(long i = 0; i < numberOfGenes ; i++){
      vector[i] = thetaAlleles[i];
    }
    ms->sort(vector,0,numberOfGenes - 1);
    ms->calcMoments(vector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
    cout << "Mean: " << average << endl;
    writeFileMainResults2 << "Mean: " << average << endl;
    cout << confidence * 100.0 << "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    cout << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
    writeFileMainResults2 << confidence * 100.0<< "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    writeFileMainResults2 << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

    // Output of Delta
    cout << "\nDelta:" << endl;
    writeFileMainResults2 << "\nDelta:" << endl;
    for(long i = 0; i < numberOfGenes ; i++){
      vector[i] = Delta[i];
    }
    ms->sort(vector,0,numberOfGenes - 1);
    ms->calcMoments(vector,numberOfGenes,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
    cout << "Mean: " << average << endl;
    writeFileMainResults2 << "Mean: " << average << endl;
    cout << confidence * 100.0 << "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    cout << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
    writeFileMainResults2 << confidence * 100.0<< "%:  (" << vector[long(numberOfGenes * ((1.0 - confidence) / 2.0))] << " , ";
    writeFileMainResults2 << vector[long(numberOfGenes * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

    cout << "\nLower 95% limit from null-distribution: " << lowerLimit << endl;
    cout << "Upper 95% limit from null-distribution: " << upperLimit << endl;
    cout << "\nNumber of directional selected genes (Delta < 0): " << numberOfPositiveSelectedGenes << " (" << (double(numberOfPositiveSelectedGenes) / double(numberOfGenes)) * 100.0 << "%)" << endl;
    cout << "Number of balancing selected genes (Delta > 0):   " << numberOfNegativeSelectedGenes << " (" << (double(numberOfNegativeSelectedGenes) / double(numberOfGenes)) * 100.0 << "%)"<< endl;
    cout << "\n----------------------------------------------------------------" << endl;
    writeFileMainResults2 << "\nLower 95% limit from null-distribution: " << lowerLimit << endl;
    writeFileMainResults2 << "Upper 95% limit from null-distribution: " << upperLimit << endl;
    writeFileMainResults2 << "\nNumber of directional selected genes (Delta < 0): " << numberOfPositiveSelectedGenes << " (" << (double(numberOfPositiveSelectedGenes) / double(numberOfGenes)) * 100.0 << "%)" << endl;
    writeFileMainResults2 << "Number of balancing selected genes (Delta > 0):   " << numberOfNegativeSelectedGenes << " (" << (double(numberOfNegativeSelectedGenes) / double(numberOfGenes)) * 100.0 << "%)"<< endl;
    writeFileMainResults2 << "\n----------------------------------------------------------------" << endl;
    writeFileMainResults2.close();

    delete alleles;
    delete geneVariance;
    delete thetaAlleles;
    delete thetaVar;
    delete Delta;
    delete vector;
  }

  cout << "\nGenerated coalescent trees were written to \"_treefile.txt\"." << endl;
  cout << "Simulated gene expression data was written to \"_datatable_synthetic_microarray.txt\"." << endl;
  if(programParameters->getMode() == 2) cout << "Test results were written to \"_tajima-type-test_results.txt\"." << endl;
  cout << "Null-distribution was written to \"_tajima-type-test_Delta-dist_neutral-case.txt\"." << endl;
  if(programParameters->getMode() == 2)   cout << "Delta-distribution of real data was written to \"_tajima-type-test_Delta-dist_real-data.txt\"." << endl;
  if(programParameters->getMode() == 2)   cout << "Probe Set IDs of directional selected genes were written to \"_tajima-type-test_directional_selected_genes.txt\"." << endl;
  if(programParameters->getMode() == 2)   cout << "Probe Set IDs of balancing selected genes were written to \"_tajima-type-test_balancing_selected_genes.txt\"." << endl;
}

// calcConfidenceLimits: Calculates the confidence limits of the null-distribution
//                        of Delta-values depending on the M&E-normal model
//                        parameters sigma_m and sigma_e.
//                        The limits are stored in *lowerLimit and *upperLimit.
void Controller::calcConfidenceLimits(double sigmaM,double sigmaE,double confidenceInterval,double *lowerLimit,double *upperLimit){

  // Initialisation
  char* inputFile = "_treefile.txt"; // name of the treefile generated by ms
  char* outputFileD = "_output_tajima-type-test_Delta-dist_neutral-case.txt";
  char charbuffer;
  char stringbuffer[4096];
  long numberOfTrees;
  long numberOfLeaves;
  bool booleanBuffer = false;
  double theta = 0.0;
  double buffer = 0.0;
  double distanceToLastExpressionValue = 0.0;
  double distanceToLastAllele = 0.0;
  double meanOfSegsites = 0.0;
  double meanOfMeanPairwiseDistance = 0.0;
  double meanOfAlleles = 0.0;
  double meanOfGeneVariance = 0.0;
  double confidence;
  BinaryTree** tree;
  ThetaAllelesEstimator* se = new ThetaAllelesEstimator();

  // Check number of trees in the file and allocate the memory
  numberOfTrees = 0;
  ifstream checkFile(inputFile);
  while(checkFile.get(charbuffer)){
    if(charbuffer == ';') numberOfTrees++;
  }
  checkFile.close();
  tree = new BinaryTree*[numberOfTrees];
  for(i = 0; i < numberOfTrees; i++){
    tree[i] = new BinaryTree();
  }

  // Initialize arrays corresponding to the treefile
  double* segsites = new double[numberOfTrees];
  double* meanPairwiseDistance = new double[numberOfTrees];
  double* alleles = new double[numberOfTrees];
  double* geneVariance = new double[numberOfTrees];
  double* theta_from_alleles = new double[numberOfTrees];
  double* theta_from_variance = new double[numberOfTrees];
  double* Delta = new double[numberOfTrees];
  double* vector = new double[numberOfTrees];

  // Read the treefile and parses all information to binarytree-objects
  ifstream file(inputFile);
  file.get(stringbuffer,4096,'/');
  file.get(charbuffer);
  file.get(stringbuffer,4096,'t');
  file.get(charbuffer);
  file.get(charbuffer);
  file.get(stringbuffer,4096,' ');
  theta = atof(stringbuffer);
  file.get(charbuffer);
  // Go on to the first tree description
  file.get(stringbuffer,4096,'/');
  file.get(charbuffer);
  file.get(charbuffer);
  file.get(charbuffer);
  for(i = 0; i < numberOfTrees; i++){
    // Reading the tree
    file.get(stringbuffer,4096,';');
    char* treeString = new char[strlen(stringbuffer)];
    strcpy(treeString,stringbuffer);
    tree[i]->init(treeString);
    // Reading the sequences
    file.get(stringbuffer,4096,':');
    file.get(charbuffer);
    file.get(stringbuffer,4096,'\n');
    segsites[i] = atof(stringbuffer);
    file.get(charbuffer);
    // Pairwise distances
    if(segsites[i] != 0){
      file.get(stringbuffer,4096,':');
      file.get(charbuffer);
      file.get(stringbuffer,4096,'\n');
      file.get(charbuffer);
      file.get(stringbuffer,4096,'/');
      meanPairwiseDistance[i] = calcMeanPairwiseDistance(stringbuffer,segsites[i],tree[i]->getNumberOfLeaves());
    } else {
      // Go on to the next tree description
      file.get(stringbuffer,4096,'/');
    }
    file.get(charbuffer);
    file.get(charbuffer);
    file.get(charbuffer);
  }
  file.close();

  // Initialize arrays corresponding to the number of leaves of the trees in the treefile
  numberOfLeaves = tree[0]->getNumberOfLeaves();
  double* gene = new double[numberOfLeaves];
  double* bufferArray = new double[2];

  // Simulate evolution of gene expression along the trees
  cout << "\nSimulate evolution of gene expression along trees ..." << endl;

  // The simulated data is stored in a file for further analysis with EMOGEE
  ofstream microarrayFile("_datatable_synthetic_microarray.txt");
  microarrayFile << "ProbeSetID\t";
  for(i = 0; i < numberOfLeaves; i++){
    microarrayFile << "1\t";
  }
  microarrayFile << "2\t3" << endl;

  // Perform one simulation for each tree in the treefile
  for(i = 0; i < numberOfTrees; i++){

    tree[i]->evolution(0.0,theta,sigmaM,gene,rnd);
    microarrayFile << i + 1 << "\t";
    for(k = 0; k < numberOfLeaves; k++){
      // The non-mutational effect is only calculated for the output file.
      // The internal data is ideal without non-mutational effects.
      microarrayFile << gene[k] + rnd->getNormalDistributedNumber(0.0,sigmaE) << "\t";
    }
    microarrayFile << "0\t0" << endl;

    // Analyze the generated gene
    geneVariance[i] = 0.0;
    alleles[i] = 0.0;

    // Estimate the number of alleles
    buffer = 1;
    ms->sort(gene,0,numberOfLeaves - 1);
    for(k = 0; k < numberOfLeaves - 1; k++){
      if(gene[k] != gene[k + 1]) buffer = buffer + 1;
    }
    // The maximum of the allele number is numberOfLeaves - 1.
    if(buffer >= numberOfLeaves) buffer = numberOfLeaves - 1;
    alleles[i] = alleles[i] + buffer;

    // Estimate the gene variance (simulate non-mutational effect)
    for(k = 0; k < numberOfLeaves; k++){
      // Add an non-mutational effect to estimate the gene variance.
      // The number of alleles has been estimated before with ideal data.
      gene[k] = gene[k] + rnd->getNormalDistributedNumber(0.0,sigmaE);
    }
    ms->calcMoments(gene,numberOfLeaves,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
    geneVariance[i] = geneVariance[i] + variance;
  }

  microarrayFile.close();

  // Calculate the mean values
  ms->calcMoments(segsites,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  meanOfSegsites = average;
  ms->calcMoments(meanPairwiseDistance,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  meanOfMeanPairwiseDistance = average;
  ms->calcMoments(alleles,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  meanOfAlleles = average;
  ms->calcMoments(geneVariance,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  meanOfGeneVariance = average;

  // Output
  ofstream writeFileMainResults("_output_tajima-type-test_main-results.txt");
  cout << "\n----------------------------------------------------------------" << endl;
  writeFileMainResults << "\n----------------------------------------------------------------" << endl;
  cout << "\nRESULTS FOR THE NEUTRAL CASE" << endl;
  writeFileMainResults << "\nRESULTS FOR THE NEUTRAL CASE" << endl;

  // Output of the gene variance
  cout << "\nGene variance:" << endl;
  writeFileMainResults << "\nGene variance:" << endl;
  for(long i = 0; i < numberOfTrees ; i++){
    vector[i] = geneVariance[i];
  }
  ms->sort(vector,0,numberOfTrees - 1);
  ms->calcMoments(vector,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  cout << "Mean: " << average << endl;
  writeFileMainResults << "Mean: " << average << endl;
  confidence = confidenceInterval;
  cout << confidence * 100.0 << "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  cout << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
  writeFileMainResults << confidence * 100.0<< "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  writeFileMainResults << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

  // Output of the number of alleles
  cout << "\nNumber of alleles:" << endl;
  writeFileMainResults << "\nNumber of alleles:" << endl;
  for(long i = 0; i < numberOfTrees ; i++){
    vector[i] = alleles[i];
  }
  ms->sort(vector,0,numberOfTrees - 1);
  ms->calcMoments(vector,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  cout << "Mean: " << average << endl;
  writeFileMainResults << "Mean: " << average << endl;
  confidence = confidenceInterval;
  cout << confidence * 100.0 << "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  cout << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
  writeFileMainResults << confidence * 100.0<< "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  writeFileMainResults << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

  // Output of theta_var
  cout << "\ntheta_var:" << endl;
  writeFileMainResults << "\ntheta_var:" << endl;
  for(long i = 0; i < numberOfTrees ; i++){
    theta_from_variance[i] = (2.0 * geneVariance[i] - 2.0 * pow(sigmaE,2.0)) / pow(sigmaM,2.0);
    vector[i] = theta_from_variance[i];
  }
  ms->sort(vector,0,numberOfTrees - 1);
  ms->calcMoments(vector,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  cout << "Mean: " << average << endl;
  writeFileMainResults << "Mean: " << average << endl;
  confidence = confidenceInterval;
  cout << confidence * 100.0 << "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  cout << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
  writeFileMainResults << confidence * 100.0<< "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  writeFileMainResults << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

  // Output of theta_alleles
  cout << "\ntheta_alleles:" << endl;
  writeFileMainResults << "\ntheta_alleles:" << endl;
  for(long i = 0; i < numberOfTrees ; i++){
    theta_from_alleles[i] = se->estimateThetaFromNumberOfAlleles(numberOfLeaves,alleles[i]);
    vector[i] = theta_from_alleles[i];
  }
  ms->sort(vector,0,numberOfTrees - 1);
  ms->calcMoments(vector,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  cout << "Mean: " << average << endl;
  writeFileMainResults << "Mean: " << average << endl;
  confidence = confidenceInterval;
  cout << confidence * 100.0 << "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  cout << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
  writeFileMainResults << confidence * 100.0<< "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  writeFileMainResults << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

  // Output of Delta
  cout << "\nDelta:" << endl;
  writeFileMainResults << "\nDelta:" << endl;
  for(long i = 0; i < numberOfTrees ; i++){
    Delta[i] = theta_from_variance[i] - theta_from_alleles[i];
    vector[i] = Delta[i];
  }
  ms->sort(vector,0,numberOfTrees - 1);
  ms->calcMoments(vector,numberOfTrees,&average,&averageDeviation,&standardDeviation,&variance,&skewness,&kurtosis);
  cout << "Mean: " << average << endl;
  writeFileMainResults << "Mean: " << average << endl;
  confidence = confidenceInterval;
  cout << confidence * 100.0 << "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  cout << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;
  writeFileMainResults << confidence * 100.0<< "%:  (" << vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))] << " , ";
  writeFileMainResults << vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))] << ")" << endl;

  cout << "\n----------------------------------------------------------------" << endl;
  writeFileMainResults << "\n----------------------------------------------------------------" << endl;

  writeFileMainResults.close();

  // Calculate confidence limits of the distribution of Delta-values
  *lowerLimit = vector[long(numberOfTrees * ((1.0 - confidence) / 2.0))];
  *upperLimit = vector[long(numberOfTrees * (confidence + ((1.0 - confidence) / 2.0)))];

  // Write the distribution of Delta-values to a file
  ofstream writeFileD(outputFileD);
  for(long i = 0; i < numberOfTrees; i++){
    writeFileD << Delta[i] << endl;
  }
  writeFileD.close();

  delete tree;
  delete se;
  delete segsites;
  delete meanPairwiseDistance;
  delete alleles;
  delete geneVariance;
  delete theta_from_alleles;
  delete theta_from_variance;
  delete Delta;
  delete vector;
  delete gene;
  delete bufferArray;
}

// calcMeanPairwiseDistance: Calculates the mean pairwise distance between a number of strings.
// positionString contains all 0/1-strings (concatenated) generated by ms (R. Hudson).
// segsites is the number of segregating sites of the strings in positionString.
// numberOfLeaves is the number of sequences in positionString which is the number of leaves in the coalescent tree.
double Controller::calcMeanPairwiseDistance(char* positionString,long segsites,long numberOfLeaves){

  // Initialisation
  double meanPairwiseDistance = 0.0;
  long i,j,k;

  // Count the differences and calculate the mean value
  for(i = 0; i < numberOfLeaves - 1; i++){
    for(j = i + 1; j < numberOfLeaves; j++){
      for(k = 0; k < segsites; k++){
        if(positionString[((segsites + 1) * i) + k] != positionString[((segsites + 1) * j) + k]) meanPairwiseDistance++;
      }
    }
  }
  meanPairwiseDistance = meanPairwiseDistance / ((numberOfLeaves * (numberOfLeaves - 1.0)) / 2.0);

  return meanPairwiseDistance;
}
