/***************************************************************************
                          data.cpp  -  description
                             -------------------
    begin                : Thu Feb 17 2005
    copyright            : (C) 2005 by Michael Rosskopf
    email                : rosskopf@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "data.h"

// Constructors and destructors
Data::Data(){
  numberOfSamples = 0;
  numberOfIndividuals = 0;
  numberOfGenes = 0;
  genes = 0;
  sampleSize = 0;
  sampleStart = 0;
  labels = 0;
  probeSetID = 0;
}
Data::~Data(){
  if(genes != 0) delete genes;
  if(sampleSize != 0) delete sampleSize;
  if(sampleStart != 0) delete sampleStart;
  if(labels != 0) delete labels;
  if(probeSetID != 0){
    for(long i = 0; i < numberOfGenes; i++){
      if(probeSetID[i] != 0) delete probeSetID[i];
    }
    delete probeSetID;
  }
}

// get- and set-methods
void Data::init(long _numberOfIndividuals,long _numberOfGenes){
  numberOfSamples = 1;
  numberOfIndividuals = _numberOfIndividuals;
  numberOfGenes = _numberOfGenes;
  labels = new long[numberOfIndividuals];
  genes = new double[numberOfIndividuals * numberOfGenes];
  probeSetID = new char*[numberOfGenes];
}
void Data::init(long _numberOfIndividuals,long _numberOfGenes,long _numberOfSamples){
  numberOfSamples = _numberOfSamples;
  numberOfIndividuals = _numberOfIndividuals;
  numberOfGenes = _numberOfGenes;
  labels = new long[numberOfIndividuals];
  sampleSize = new long[_numberOfSamples];
  sampleStart = new long[_numberOfSamples];
  genes = new double[numberOfIndividuals * numberOfGenes];
  probeSetID = new char*[numberOfGenes];
}
long Data::getNumberOfSamples(){
  return numberOfSamples;
}
long Data::getNumberOfIndividuals(){
  return numberOfIndividuals;
}
long Data::getNumberOfGenes(){
  return numberOfGenes;
}
long Data::getLabel(long _sample){
  return labels[_sample];
}
long Data::getSampleSize(long _sampleSize){
  return sampleSize[_sampleSize];
}
long Data::getSampleStart(long _sampleNumber){
  return sampleStart[_sampleNumber];
}
char* Data::getProbeSetID(long index){
  return probeSetID[index];
}
double Data::getGene(long _sample,long _gene){
  return genes[(numberOfIndividuals * _gene) + _sample];
}
void Data::setLabel(long index,long _label){
  labels[index] = _label;
}
void Data::setSampleSize(long _sampleNumber,long _value){
  sampleSize[_sampleNumber] = _value;
}
void Data::setSampleStart(long _sampleNumber,long _value){
  sampleStart[_sampleNumber] = _value;
}
void Data::setGene(long _sample,long _gene,double expressionValue){
  genes[(numberOfIndividuals * _gene) + _sample] = expressionValue;
}

// toString
void Data::toString(){
  cout << "ProbeSetID\t";
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    cout << labels[i] << "\t";
  }
  cout << labels[numberOfIndividuals - 1] << "\n";
  for(long i = 0; i < numberOfGenes; i++){
  cout << i + 1 << "\t";
    for(long j = 0; j < numberOfIndividuals - 1; j++){
      cout << genes[(numberOfIndividuals * i) + j] << "\t";
    }
    cout << genes[(numberOfIndividuals * i) + numberOfIndividuals - 1] << "\n";
  }
}

// dataCheck: Checks whether sample labels are in the data set
bool Data::dataCheck(ProgramParameters* programParameters){
  bool globalvalid = true;
  bool valid;
  // Checks whether sample1 is inside the data set
  valid = false;
  for(long i = 0; i < numberOfIndividuals; i++){
    if(programParameters->getSample1() == labels[i]) valid = true;
  }
  if(!valid){
    cout << "\nError: Data set does not contain sample1 = " << programParameters->getSample1() << "." << endl;
    globalvalid = false;
  }
  // Checks whether sample2 is inside the data set if tool 1 is used (Bayesian method)
  // sample2 is not used by the Tajima-type test
  if(programParameters->getTool() == 1){
    valid = false;
    for(long i = 0; i < numberOfIndividuals; i++){
      if(programParameters->getSample2() == labels[i]) valid = true;
    }
    if(!valid){
      cout << "\nError: Data set does not contain sample2 = " << programParameters->getSample2() << "." << endl;
      globalvalid = false;
    }
  }
  return globalvalid;
}

// calcGeneVariance: Calculates the variance of a gene in a specific sample of individuals
double Data::calcGeneVariance(long sample,long gene){
  MathFunctions ms = MathFunctions();
  double ave;
  double adev;
  double sdev;
  double var;
  double skew;
  double curt;
  double* geneVector = new double[sampleSize[sample - 1]];
  double geneVariance = 0.0;

  // Calc the variance of a gene in a group
  for(long i = 0; i < sampleSize[sample - 1]; i++){
    geneVector[i] = getGene(sampleStart[sample - 1] + i,gene);
  }
  ms.calcMoments(geneVector,sampleSize[sample - 1],&ave,&adev,&sdev,&var,&skew,&curt);
  geneVariance = var;

  delete geneVector;
  return geneVariance;
}

// calcNumberOfAlleles:
// Calculates the number of alleles for a gene in a sample of individuals.
// sigmaM and sigmaE are model parameters.
double Data::calcNumberOfAlleles(long group, long gene, double sigmaM, double sigmaE){

  MathFunctions ms = MathFunctions();
  double ave;
  double adev;
  double sdev;
  double var;
  double skew;
  double curt;
  double* geneVector = new double[sampleSize[group - 1]];
  double numberOfAlleles = 1;
  double distanceToLastExpressionValue = 0.0;
  double distanceToLastAllele = 0.0;

  // Sort expression values
  for(long i = 0; i < sampleSize[group - 1]; i++){
    geneVector[i] = getGene(sampleStart[group - 1] + i,gene);
  }
  ms.sort(geneVector,0,sampleSize[group - 1] - 1);

  // Estimate number of alleles
  for(long i = 0; i < sampleSize[group - 1] - 1; i++){
    distanceToLastExpressionValue = geneVector[i + 1] - geneVector[i];
    distanceToLastAllele = distanceToLastAllele + distanceToLastExpressionValue;
    if((distanceToLastExpressionValue > (2.0 * sigmaE)) || (distanceToLastAllele > (2.0 * sigmaM))){
      // Count a new allele
      numberOfAlleles++;
      distanceToLastAllele = 0.0;
    }
  }

  // The maximum of the allele number is the sampleSize - 1.
  if(numberOfAlleles >= sampleSize[group - 1]) numberOfAlleles = sampleSize[group - 1] - 1;

  return numberOfAlleles;
}

// checkDataSize: Called from the load-method. This method is used
//                to check the data size and to initialize the data structure.
void Data::checkDataSize(char* filename){
  char charbuffer;
  numberOfIndividuals = 0;
  numberOfGenes = 0;
  ifstream file(filename);
  while(file.get(charbuffer)){
    if(charbuffer == '\n') break;
    if(charbuffer == '\t') numberOfIndividuals++;
  }
  while(file.get(charbuffer)){
    if(charbuffer == '\n') numberOfGenes++;
  }
  file.close();
  init(numberOfIndividuals,numberOfGenes);
}

// load: Loads a dataset. The initialization occurs automatically
void Data::load(char* filename){
  checkDataSize(filename);
  char charbuffer;
  char stringbuffer[64];
  ifstream file(filename);
  file.get(stringbuffer,64,'\t');
  file.get(charbuffer);
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    file.get(stringbuffer,64,'\t');
    labels[i] = atol(stringbuffer);
    file.get(charbuffer);
  }
  file.get(stringbuffer,64,'\n');
  labels[numberOfIndividuals - 1] = atol(stringbuffer);
  file.get(charbuffer);

  for(long i = 0; i < numberOfGenes; i++){
  file.get(stringbuffer,64,'\t');
  probeSetID[i] = new char[sizeof(stringbuffer)];
  strcpy(probeSetID[i],stringbuffer);
  file.get(charbuffer);
    for(long j = 0; j < numberOfIndividuals - 1; j++){
      file.get(stringbuffer,64,'\t');
      genes[(numberOfIndividuals * i) + j] = atof(stringbuffer);
      file.get(charbuffer);
    }
    file.get(stringbuffer,64,'\n');
    genes[(numberOfIndividuals * i) + numberOfIndividuals - 1] = atof(stringbuffer);
    file.get(charbuffer);
  }
  file.close();
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    if(labels[i] != labels[i + 1]) numberOfSamples++;
  }
  sampleSize = new long[numberOfSamples];
  sampleStart = new long[numberOfSamples];
  for(long i = 0; i < numberOfSamples; i ++){
    sampleSize[i] = 0;
    sampleStart[i] = 0;
  }
  long groupCounter = 1;
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    if(labels[i] != labels[i + 1]){
      sampleStart[groupCounter] = i + 1;
      groupCounter++;
    }
    sampleSize[labels[i] - 1]++;
  }
  sampleSize[labels[numberOfIndividuals - 1] - 1]++;
}

// save: Saves a dataset.
void Data::save(char* filename){
  ofstream file(filename);
  file.precision(10);
  file << "ProbeSetID\t";
  for(long i = 0; i < numberOfIndividuals - 1; i++){
    file << labels[i] << "\t";
  }
  file << labels[numberOfIndividuals - 1] << "\n";
  for(long i = 0; i < numberOfGenes; i++){
    file << i + 1 << "\t";
    for(long j = 0; j < numberOfIndividuals - 1; j++){
      file << genes[(numberOfIndividuals * i) + j] << "\t";
    }
    file << genes[(numberOfIndividuals * i) + numberOfIndividuals - 1] << "\n";
  }
  file.close();
}
